import numpy as np
import pytest
import xarray as xr

import xugrid as xu
from xugrid.ugrid import partitioning as pt


def generate_mesh_2d(nx, ny, name="mesh2d"):
    points = [
        (x, y) for y in np.linspace(0, ny, ny + 1) for x in np.linspace(0, nx, nx + 1)
    ]
    connectivity = [
        [
            it + jt * (nx + 1),
            it + jt * (nx + 1) + 1,
            it + (jt + 1) * (nx + 1) + 1,
            it + (jt + 1) * (nx + 1),
        ]
        for jt in range(ny)
        for it in range(nx)
    ]

    return xu.Ugrid2d(*np.array(points).T, -1, np.array(connectivity), name=name)


def test_labels_to_indices():
    labels = np.array([0, 1, 0, 2, 2])
    indices = pt.labels_to_indices(labels)
    assert np.array_equal(indices[0], [0, 2])
    assert np.array_equal(indices[1], [1])
    assert np.array_equal(indices[2], [3, 4])


class TestGridPartitioning:
    @pytest.fixture(autouse=True)
    def setup(self):
        """
        Generate simple 2D mesh connectivity with rectangular elements, eg:

              10 -- 11 -- 12 -- 13 -- 14
              |     |     |     |     |
              5 --- 6 --- 7 --- 8 --- 9
              |     |     |     |     |
              0 --- 1 --- 2 --- 3 --- 4
        """
        self.grid = generate_mesh_2d(5, 3)

    def test_label_partition(self):
        labels = self.grid.label_partitions(n_part=3)
        assert isinstance(labels, xu.UgridDataArray)
        assert labels.name == "labels"
        assert labels.ugrid.grid == self.grid
        assert np.allclose(np.unique(labels.values), [0, 1, 2])

    def test_partition(self):
        parts = self.grid.partition(n_part=3)
        assert len(parts) == 3
        for part in parts:
            assert isinstance(part, xu.Ugrid2d)


class TestDatasetPartition:
    @pytest.fixture(autouse=True)
    def setup(self):
        grid = generate_mesh_2d(5, 3)
        ds = xr.Dataset()
        ds["node_z"] = xr.DataArray(np.arange(grid.n_node), dims=(grid.node_dimension,))
        ds["edge_z"] = xr.DataArray(np.arange(grid.n_edge), dims=(grid.edge_dimension,))
        ds["face_z"] = xr.DataArray(np.arange(grid.n_face), dims=(grid.face_dimension,))
        uds = xu.UgridDataset(obj=ds, grids=[grid])
        self.uds = uds
        self.grid = grid
        self.obj = self.uds.ugrid.obj
        self.labels = xu.UgridDataArray(
            xr.DataArray(
                data=np.repeat([0, 1, 2], 5),
                dims=(self.grid.face_dimension,),
            ),
            self.grid,
        )

    def test_partition_by_labels__errors(self):
        with pytest.raises(TypeError, match="labels must be a UgridDataArray"):
            pt.partition_by_label(self.grid, self.obj, np.arange(15))

        with pytest.raises(TypeError, match="labels must have integer dtype"):
            pt.partition_by_label(self.grid, self.obj, self.labels.astype(float))

        other_grid = generate_mesh_2d(3, 3)
        with pytest.raises(
            ValueError, match="grid of labels does not match xugrid object"
        ):
            pt.partition_by_label(other_grid, self.obj, self.labels)

        dim_labels = self.labels.expand_dims("somedim", axis=0)
        with pytest.raises(ValueError, match="Can only partition this topology"):
            pt.partition_by_label(self.grid, self.obj, dim_labels)

        with pytest.raises(TypeError, match="Expected DataArray or Dataset"):
            pt.partition_by_label(self.grid, self.obj.values, self.labels)

    def test_partition_by_labels__dataset(self):
        partitions = pt.partition_by_label(self.grid, self.obj, self.labels)
        assert len(partitions) == 3
        for partition in partitions:
            assert isinstance(partition, xu.UgridDataset)
            assert "face_z" in partition
            assert "edge_z" in partition
            assert "node_z" in partition

    def test_partition_by_labels__dataarray(self):
        partitions = pt.partition_by_label(self.grid, self.obj["face_z"], self.labels)
        assert len(partitions) == 3
        for partition in partitions:
            assert isinstance(partition, xu.UgridDataArray)
            assert partition.name == "face_z"

    def test_partition_roundtrip(self):
        partitions = self.uds.ugrid.partition(n_part=4)
        back = pt.merge_partitions(partitions)
        assert isinstance(back, xu.UgridDataset)

        order = np.argsort(back["face_z"].values)
        reordered = back.isel(mesh2d_nFaces=order)
        assert reordered["face_z"].equals(self.uds["face_z"])

    def test_merge_partitions__errors(self):
        partitions = self.uds.ugrid.partition(n_part=2)
        with pytest.raises(TypeError, match="Expected UgridDataArray or UgridDataset"):
            pt.merge_partitions(p.ugrid.obj for p in partitions)

        grid1 = partitions[1].ugrid.grid
        partitions[1]["extra"] = (grid1.face_dimension, np.ones(grid1.n_face))
        with pytest.raises(ValueError, match="These variables are present"):
            pt.merge_partitions(partitions)

        partitions = self.uds.ugrid.partition(n_part=2)
        partitions[1]["face_z"] = partitions[1]["face_z"].expand_dims("layer", axis=0)
        with pytest.raises(ValueError, match="Dimensions for face_z do not match"):
            pt.merge_partitions(partitions)

        uds = self.uds.copy()
        grid = uds.ugrid.grid
        uds["two_dim"] = (
            ("mesh2d_nNodes", "mesh2d_nEdges"),
            np.ones((grid.n_node, grid.n_edge)),
        )
        partitions = uds.ugrid.partition(n_part=2)
        with pytest.raises(
            ValueError, match="two_dim contains more than one UGRID dimension"
        ):
            pt.merge_partitions(partitions)

    def test_merge_partitions_no_duplicates(self):
        part1 = self.uds.isel(mesh2d_nFaces=[0, 1, 2, 3])
        part2 = self.uds.isel(mesh2d_nFaces=[2, 3, 4, 5])
        merged = pt.merge_partitions([part1, part2])
        assert np.bincount(merged["face_z"] == 1).all()


class TestMultiTopologyMergePartitions:
    @pytest.fixture(autouse=True)
    def setup(self):
        grid_a = generate_mesh_2d(2, 3, "first")
        grid_b = generate_mesh_2d(4, 5, "second")
        parts_a = grid_a.partition(n_part=2)
        parts_b = grid_b.partition(n_part=2)

        datasets = []
        for i, (part_a, part_b) in enumerate(zip(parts_a, parts_b)):
            ds = xu.UgridDataset(grids=[part_a, part_b])
            ds["a"] = ((part_a.face_dimension), np.arange(part_a.n_face))
            ds["b"] = ((part_b.face_dimension), np.arange(part_b.n_face))
            ds["c"] = i
            datasets.append(ds)

        self.datasets = datasets

    def test_merge_partitions(self):
        merged = pt.merge_partitions(self.datasets)
        assert isinstance(merged, xu.UgridDataset)
        assert len(merged.ugrid.grids) == 2
        # In case of non-UGRID data, it should default to the first partition:
        assert merged["c"] == 0

    def test_merge_partitions__errors(self):
        pa = self.datasets[0][["a"]]
        pb = self.datasets[1][["b"]]
        with pytest.raises(ValueError, match="Expected 2 UGRID topologies"):
            pt.merge_partitions([pa, pb])

        grid_a = self.datasets[1].ugrid.grids[0].copy()
        grid_c = self.datasets[1].ugrid.grids[1].copy()
        grid_c._attrs["face_dimension"] = "abcdef"
        dataset2 = xu.UgridDataset(
            obj=self.datasets[1].ugrid.obj, grids=[grid_a, grid_c]
        )
        with pytest.raises(ValueError, match="Dimension names on UGRID topology"):
            pt.merge_partitions([self.datasets[0], dataset2])

        xy = np.array(
            [
                [0.0, 0.0],
                [1.0, 1.0],
                [2.0, 2.0],
            ]
        )
        grid_d = xu.Ugrid1d(
            node_x=xy[:, 0],
            node_y=xy[:, 1],
            fill_value=-1,
            edge_node_connectivity=np.array([[0, 1], [1, 2]]),
            name="second",
        )
        dataset3 = xu.UgridDataset(
            obj=self.datasets[1].ugrid.obj, grids=[grid_a, grid_d]
        )
        with pytest.raises(
            TypeError, match="All partition topologies with name second"
        ):
            pt.merge_partitions([self.datasets[0], dataset3])



from xugrid.data.sample_data import adh_san_diego, elevation_nl, provinces_nl, xoxo
from xugrid.data.synthetic import disk, generate_disk

__all__ = (
    "adh_san_diego",
    "elevation_nl",
    "provinces_nl",
    "xoxo",
    "disk",
    "generate_disk",
)



"""
Wrap in advance instead of overloading __getattr__.

This allows for tab completion and documentation.
"""

import types
from collections import ChainMap
from functools import wraps
from itertools import chain
from typing import List, Sequence, Union

import xarray as xr
from pandas import RangeIndex

import xugrid
from xugrid.conversion import grid_from_dataset, grid_from_geodataframe
from xugrid.ugrid.ugrid2d import Ugrid2d
from xugrid.ugrid.ugridbase import AbstractUgrid, UgridType, align

# Import entire module here for circular import of UgridDatasetAccessor and
# UgridDataArrayAccessor. Note: can only be used in functions (since that code
# is run at runtime).


def maybe_xugrid(obj, topology, old_indexes=None):
    if not isinstance(obj, (xr.DataArray, xr.Dataset)):
        return obj

    # Topology can either be a sequence of grids or a grid.
    if isinstance(topology, (list, set, tuple)):
        grids = {dim: grid for grid in topology for dim in grid.dimensions}
    else:
        grids = {dim: topology for dim in topology.dimensions}

    item_grids = list({grids[dim] for dim in obj.dims if dim in grids})

    if len(item_grids) == 0:
        return obj
    else:
        result, aligned = align(obj, item_grids, old_indexes)

        if isinstance(result, xr.DataArray):
            if len(aligned) > 1:
                raise RuntimeError("This shouldn't happen. Please open an issue.")
            return UgridDataArray(result, aligned[0])

        elif isinstance(result, xr.Dataset):
            return UgridDataset(result, aligned)


def maybe_xarray(arg):
    if isinstance(arg, (UgridDataArray, UgridDataset)):
        return arg.obj
    else:
        return arg


def wraps_xarray(method):
    @wraps(method)
    def wrapped(*args, **kwargs):
        self = args[0]
        args = [maybe_xarray(arg) for arg in args]
        kwargs = {k: maybe_xarray(v) for k, v in kwargs.items()}
        result = method(*args, **kwargs)

        # Sidestep staticmethods, classmethods: in that case self will not be a
        # xugrid type.
        if isinstance(self, (UgridDataArray, UgridDataset)):
            return maybe_xugrid(result, self.grids, self.obj.indexes)
        else:
            return result

    return wrapped


def wrap_accessor(accessor):
    # TODO: This will not add dynamic accessors, those most be included at
    # runtime instead?

    def wrapped(*args, **kwargs):
        args = [maybe_xarray(arg) for arg in args]
        kwargs = {k: maybe_xarray(v) for k, v in kwargs.items()}
        result = accessor(*args, **kwargs)
        return result

    return wrapped


def wrap(
    target_class_dict,
    source_class,
):
    FuncType = (types.FunctionType, types.MethodType)

    # Set every method, property from the xarray object to the UgridDataArray,
    # UgridDataset. Don't set everything, as this will break the objects.
    #
    # class Empty:
    #     pass
    #
    # keep = {
    #     "__eq__",
    #     "__ge__",
    #     "__gt__",
    #     "__le__",
    #     "__lt__",
    #     "__ne__",
    #     "__repr__",
    #     "__str__",
    # }
    #
    # remove = set(dir(Empty)) - keep

    remove = {
        # These members are shared by all objects:
        "__class__",
        "__delattr__",
        "__dict__",
        "__dir__",
        "__doc__",
        "__format__",
        "__getattribute__",
        "__hash__",
        "__init__",
        "__init_subclass__",
        "__module__",
        "__new__",
        "__reduce__",
        "__reduce_ex__",
        "__setattr__",
        "__sizeof__",
        "__subclasshook__",
        "__weakref__"
        # These are additionally included in xarray:
        "__getatrr__",
        "__slots__",
        "__annotations__",
    }

    attr_names = set(dir(source_class)) - remove
    all_attrs = {k: getattr(source_class, k) for k in attr_names}

    methods = {k: v for k, v in all_attrs.items() if isinstance(v, FuncType)}
    for name, method in methods.items():
        wrapped = wraps_xarray(method)
        setattr(wrapped, "__doc__", method.__doc__)
        target_class_dict[name] = wrapped

    properties = {k: v for k, v in all_attrs.items() if isinstance(v, property)}
    for name, prop in properties.items():
        wrapped = property(
            fget=wraps_xarray(prop.__get__),
            fset=wraps_xarray(prop.__set__),
            doc=prop.__doc__,
        )
        target_class_dict[name] = wrapped

    accessors = {k: v for k, v in all_attrs.items() if isinstance(v, type)}
    for name, accessor in accessors.items():
        wrapped = property(wrap_accessor(accessor))
        setattr(wrapped, "__doc__", accessor.__doc__)
        target_class_dict[name] = wrapped

    return


class DataArrayForwardMixin:
    wrap(
        target_class_dict=vars(),
        source_class=xr.DataArray,
    )


class DatasetForwardMixin:
    wrap(
        target_class_dict=vars(),
        source_class=xr.Dataset,
    )


def assign_ugrid_coords(obj, grids):
    grid_dims = ChainMap(*(grid.dimensions for grid in grids))
    ugrid_dims = set(grid_dims.keys()).intersection(obj.dims)
    ugrid_coords = {dim: RangeIndex(0, grid_dims[dim]) for dim in ugrid_dims}
    obj = obj.assign_coords(ugrid_coords)
    return obj


class UgridDataArray(DataArrayForwardMixin):
    def __init__(self, obj: xr.DataArray, grid: UgridType):
        if not isinstance(obj, xr.DataArray):
            raise TypeError(
                "obj must be xarray.DataArray. Received instead: "
                f"{type(obj).__name__}"
            )
        if not isinstance(grid, AbstractUgrid):
            raise TypeError(
                "grid must be Ugrid1d or Ugrid2d. Received instead: "
                f"{type(grid).__name__}"
            )

        self._grid = grid
        self._obj = assign_ugrid_coords(obj, [grid])

    def __getattr__(self, attr):
        result = getattr(self.obj, attr)
        return maybe_xugrid(result, [self.grid])

    @property
    def obj(self):
        return self._obj

    @property
    def grid(self):
        return self._grid

    @property
    def grids(self) -> List[UgridType]:
        return [self._grid]

    @property
    def ugrid(self):
        """
        UGRID Accessor. This "accessor" makes operations using the UGRID
        topology available.
        """
        return xugrid.core.dataarray_accessor.UgridDataArrayAccessor(
            self.obj, self.grid
        )

    @staticmethod
    def from_structured(da: xr.DataArray):
        """
        Create a UgridDataArray from a (structured) xarray DataArray.

        The spatial dimensions are flattened into a single UGRID face dimension.

        Parameters
        ----------
        da: xr.DataArray
            Last two dimensions must be ``("y", "x")``.

        Returns
        -------
        unstructured: UgridDataArray
        """
        if da.dims[-2:] != ("y", "x"):
            raise ValueError('Last two dimensions of da must be ("y", "x")')
        grid = Ugrid2d.from_structured(da)
        dims = da.dims[:-2]
        coords = {k: da.coords[k] for k in dims}
        face_da = xr.DataArray(
            da.data.reshape(*da.shape[:-2], -1),
            coords=coords,
            dims=[*dims, grid.face_dimension],
            name=da.name,
        )
        return UgridDataArray(face_da, grid)


class UgridDataset(DatasetForwardMixin):
    def __init__(
        self,
        obj: xr.Dataset = None,
        grids: Union[UgridType, Sequence[UgridType]] = None,
    ):
        if obj is None and grids is None:
            raise ValueError("At least either obj or grids is required")

        if obj is None:
            ds = xr.Dataset()
        else:
            if not isinstance(obj, xr.Dataset):
                raise TypeError(
                    "obj must be xarray.Dataset. Received instead: "
                    f"{type(obj).__name__}"
                )
            connectivity_vars = [
                name
                for v in obj.ugrid_roles.connectivity.values()
                for name in v.values()
            ]
            ds = obj.drop_vars(obj.ugrid_roles.topology + connectivity_vars)

        if grids is None:
            topologies = obj.ugrid_roles.topology
            grids = [grid_from_dataset(obj, topology) for topology in topologies]
        else:
            # Make sure it's a new list
            if isinstance(grids, (list, tuple, set)):
                grids = list(grids)
            else:  # not iterable
                grids = [grids]
            # Now typecheck
            for grid in grids:
                if not isinstance(grid, AbstractUgrid):
                    raise TypeError(
                        "grid must be Ugrid1d or Ugrid2d. "
                        f"Received instead: {type(grid).__name__}"
                    )

        self._grids = grids
        self._obj = assign_ugrid_coords(ds, grids)

    @property
    def obj(self):
        return self._obj

    @property
    def grid(self) -> UgridType:
        # We need to do some checking. Don't duplicate that logic.
        return self.ugrid.grid

    @property
    def grids(self) -> List[UgridType]:
        return self._grids

    @property
    def ugrid(self):
        """
        UGRID Accessor. This "accessor" makes operations using the UGRID
        topology available.
        """
        return xugrid.core.dataset_accessor.UgridDatasetAccessor(self.obj, self.grids)

    def __getattr__(self, attr):
        result = getattr(self.obj, attr)
        return maybe_xugrid(result, self.grids)

    def __setitem__(self, key, value):
        # TODO: check with topology
        if isinstance(value, UgridDataArray):
            append = True
            # Check if the dimensions occur in self.
            # if they don't, the grid should be added.
            if self.grids is not None:
                alldims = set(
                    chain.from_iterable([grid.dimensions for grid in self.grids])
                )
                matching_dims = set(value.grid.dimensions).intersection(alldims)
                if matching_dims:
                    append = False
                    # If they do match: the grids should match.
                    grids = {
                        dim: grid for grid in self.grids for dim in grid.dimensions
                    }
                    firstdim = next(iter(matching_dims))
                    grid_to_check = grids[firstdim]
                    if not grid_to_check.equals(value.grid):
                        raise ValueError(
                            "Grids share dimension names but do not are not identical. "
                            f"Matching dimensions: {matching_dims}"
                        )

            self.obj[key] = value.obj
            if append:
                self._grids.append(value.grid)
        else:
            self.obj[key] = value

    @staticmethod
    def from_geodataframe(geodataframe: "geopandas.GeoDataFrame"):  # type: ignore # noqa
        """
        Convert a geodataframe into the appropriate Ugrid topology and dataset.

        Parameters
        ----------
        geodataframe: gpd.GeoDataFrame

        Returns
        -------
        dataset: UGridDataset
        """
        grid = grid_from_geodataframe(geodataframe)
        ds = xr.Dataset.from_dataframe(geodataframe.drop("geometry", axis=1))
        return UgridDataset(ds, [grid])



import numpy as np
import pytest

from xugrid.regrid import reduce


@pytest.fixture(scope="function")
def args():
    values = np.array([0.0, 1.0, 2.0, np.nan, 3.0, 4.0])
    indices = np.array([0, 1, 2, 3])
    weights = np.array([0.5, 0.5, 0.5, 0.5])
    return values, indices, weights


def test_mean(args):
    actual = reduce.mean(*args)
    assert np.allclose(actual, 1.0)


def test_harmonic_mean(args):
    actual = reduce.harmonic_mean(*args)
    assert np.allclose(actual, 1.0 / (0.5 / 1.0 + 0.5 / 2.0))


def test_geometric_mean(args):
    actual = reduce.geometric_mean(*args)
    assert np.allclose(actual, np.sqrt(1.0 * 2.0))


def test_sum(args):
    actual = reduce.sum(*args)
    assert np.allclose(actual, 3.0)


def test_minimum(args):
    actual = reduce.minimum(*args)
    assert np.allclose(actual, 0.0)


def test_maximum(args):
    actual = reduce.maximum(*args)
    assert np.allclose(actual, 2.0)


def test_mode(args):
    actual = reduce.mode(*args)
    assert np.allclose(actual, 0.0)

    values = np.array([0.0, 1.0, 1.0, 2.0, np.nan])
    indices = np.array([0, 1, 2, 3, 4])
    weights = np.array([0.5, 0.5, 0.5, 0.5, 0.5])
    args = (values, indices, weights)
    actual = reduce.mode(*args)
    assert np.allclose(actual, 1.0)
    # The weights shouldn't be mutated!
    assert np.allclose(weights, 0.5)


def test_median(args):
    actual = reduce.median(*args)
    assert np.allclose(actual, 1.0)


def test_conductance(args):
    actual = reduce.conductance(*args)
    assert np.allclose(actual, 1.5)


def test_max_overlap(args):
    actual = reduce.max_overlap(*args)
    assert np.allclose(actual, 0.0)

    values = np.array([0.0, 1.0, 2.0, np.nan])
    indices = np.array([0, 1, 2, 3])
    weights = np.array([0.5, 1.5, 0.5, 2.5])
    args = (values, indices, weights)
    actual = reduce.max_overlap(*args)
    assert np.allclose(actual, 1.0)



"""
Copied from xarray.tests.test_utils.py

The reason is that the content of xarray.core.utils are all private methods.
Hence, Xarray provides no guarantees on breaking changes.

Xarray is licensed under Apache License 2.0:
https://github.com/pydata/xarray/blob/main/LICENSE
"""
import pytest

from xugrid.core.utils import either_dict_or_kwargs


def test_either_dict_or_kwargs():
    result = either_dict_or_kwargs({"a": 1}, None, "foo")
    expected = {"a": 1}
    assert result == expected

    result = either_dict_or_kwargs(None, {"a": 1}, "foo")
    expected = {"a": 1}
    assert result == expected

    with pytest.raises(ValueError, match=r"foo"):
        result = either_dict_or_kwargs({"a": 1}, {"a": 1}, "foo")



import abc
import copy
from itertools import chain
from typing import Tuple, Type, Union

import numpy as np
import pandas as pd
import xarray as xr
from scipy.sparse import csr_matrix

from xugrid.constants import BoolArray, FloatArray, IntArray
from xugrid.ugrid import connectivity, conventions


def as_pandas_index(index: Union[BoolArray, IntArray, pd.Index], n: int):
    if isinstance(index, np.ndarray):
        if index.size > n:
            raise ValueError(
                f"index size {index.size} is larger than dimension size: {n}"
            )
        if np.issubdtype(index.dtype, np.bool_):
            # Significantly quicker if all true.
            if index.all():
                pd_index = pd.RangeIndex(0, n)
            else:
                pd_index = pd.Index((np.arange(n)[index]))
        elif np.issubdtype(index.dtype, np.integer):
            pd_index = pd.Index(index)
        else:
            raise TypeError(f"index should be bool or integer. Received: {index.dtype}")

    elif isinstance(index, pd.Index):
        pd_index = index

    else:
        raise TypeError(
            "index should be pandas Index or numpy array. Received: "
            f"{type(index).__name__}"
        )

    if not pd_index.is_unique:
        raise ValueError(
            "index contains repeated values; only subsets will result "
            "in valid UGRID topology."
        )
    #    # TODO?
    #    # Uniqueness is required, but sorting arguably not.
    #    if not pd_index.is_monotonic_increasing:
    #        raise NotImplementedError("UGRID indexes must be sorted and unique.")

    return pd_index


def align(obj, grids, old_indexes):
    """
    Check which indexes have changed. Index on those new values.
    If none are changed, return (obj, grids) as is.
    """
    if old_indexes is None:
        return obj, grids

    ugrid_dims = set(
        chain.from_iterable(grid.dimensions for grid in grids)
    ).intersection(old_indexes)
    new_indexes = {
        k: index
        for k, index in obj.indexes.items()
        if (k in ugrid_dims) and (not index.equals(old_indexes[k]))
    }
    if not new_indexes:
        return obj, grids

    # Group the indexers by grid
    new_grids = []
    for grid in grids:
        ugrid_dims = set(grid.dimensions).intersection(new_indexes)
        ugrid_indexes = {dim: new_indexes[dim] for dim in ugrid_dims}
        newgrid, indexers = grid.isel(indexers=ugrid_indexes, return_index=True)
        indexers = {
            k: v for k, v in indexers.items() if k in obj.dims and k not in new_indexes
        }
        obj = obj.isel(indexers)
        new_grids.append(newgrid)
    return obj, new_grids


class AbstractUgrid(abc.ABC):
    @abc.abstractproperty
    def topology_dimension():
        pass

    @abc.abstractproperty
    def core_dimension():
        pass

    @abc.abstractproperty
    def dimensions():
        pass

    @abc.abstractproperty
    def mesh():
        pass

    @abc.abstractproperty
    def meshkernel():
        pass

    @abc.abstractstaticmethod
    def from_dataset():
        pass

    @abc.abstractmethod
    def to_dataset():
        pass

    @abc.abstractmethod
    def topology_subset():
        pass

    @abc.abstractmethod
    def clip_box():
        pass

    @abc.abstractmethod
    def sel_points():
        pass

    @abc.abstractmethod
    def intersect_line():
        pass

    @abc.abstractmethod
    def intersect_linestring():
        pass

    @abc.abstractmethod
    def sel():
        pass

    @abc.abstractmethod
    def _clear_geometry_properties():
        pass

    @abc.abstractstaticmethod
    def merge_partitions():
        pass

    @abc.abstractmethod
    def reindex_like():
        pass

    @abc.abstractmethod
    def connectivity_matrix(self, dim: str, xy_weights: bool) -> csr_matrix:
        pass

    def _initialize_indexes_attrs(self, name, dataset, indexes, attrs):
        defaults = conventions.default_topology_attrs(name, self.topology_dimension)

        if dataset is None:
            if attrs is None:
                x, y = defaults["node_coordinates"].split()
                indexes = {"node_x": x, "node_y": y}
            else:
                if indexes is None:
                    raise ValueError("indexes must be provided for attrs")
                defaults.update(attrs)

            self._indexes = indexes
            self._attrs = defaults

        else:
            if attrs is not None:
                raise ValueError("Provide either dataset or attrs, not both.")
            if indexes is None:
                raise ValueError("indexes must be provided for dataset")

            derived_dims = dataset.ugrid_roles.dimensions[name]
            self._indexes = indexes
            self._attrs = {**defaults, **derived_dims, **dataset[name].attrs}

        # Ensure the name is always in sync.
        self._attrs["name"] = name
        return

    def rename(self, name: str, return_name_dict: bool = False):
        """
        Create a new grid with all variables named according to the default
        naming conventions.
        """
        # Get the old and the new names. Their keys are the same.
        old_attrs = self._attrs
        new_attrs = conventions.default_topology_attrs(name, self.topology_dimension)

        # The attrs will have some roles joined together, e.g. node_coordinates
        # will contain x and y as "mesh2d_node_x mesh2d_node_y".
        name_dict = {self.name: name}
        skip = ("cf_role", "long_name", "topology_dimension")
        for key, value in old_attrs.items():
            if key in new_attrs and key not in skip:
                split_new = new_attrs[key].split()
                split_old = value.split()
                if len(split_new) != len(split_old):
                    raise ValueError(
                        f"Number of entries does not match on {key}: "
                        f"{split_new} versus {split_old}"
                    )
                for name_key, name_value in zip(split_old, split_new):
                    name_dict[name_key] = name_value

        new = self.copy()
        new.name = name
        new._attrs = new_attrs
        new._indexes = {k: name_dict[v] for k, v in new._indexes.items()}
        if new._dataset is not None:
            to_rename = (
                tuple(new._dataset.data_vars)
                + tuple(new._dataset.coords)
                + tuple(new._dataset.dims)
            )
            new._dataset = new._dataset.rename(
                {k: v for k, v in name_dict.items() if k in to_rename}
            )

        if return_name_dict:
            return new, name_dict
        else:
            return new

    @staticmethod
    def _single_topology(dataset: xr.Dataset):
        topologies = dataset.ugrid_roles.topology
        n_topology = len(topologies)
        if n_topology == 0:
            raise ValueError("Dataset contains no UGRID topology variable.")
        elif n_topology > 1:
            raise ValueError(
                f"Dataset contains {n_topology} topology variables, "
                "please specify the topology variable name to use."
            )
        return topologies[0]

    def _filtered_attrs(self, dataset: xr.Dataset):
        """Remove names that are not present in the dataset."""
        topodim = self.topology_dimension
        attrs = self._attrs.copy()

        ugrid_dims = conventions._DIM_NAMES[topodim] + tuple(
            [dims[0] for dims in conventions._CONNECTIVITY_DIMS.values()]
        )
        for key in ugrid_dims:
            if key in attrs and attrs[key] not in dataset.dims:
                attrs.pop(key)

        for key in conventions._CONNECTIVITY_NAMES[topodim]:
            if key in attrs and attrs[key] not in dataset:
                attrs.pop(key)

        for coord in conventions._COORD_NAMES[topodim]:
            if coord in attrs:
                names = attrs[coord].split(" ")
                present = [name for name in names if name in dataset]
                if present:
                    attrs[coord] = " ".join(present)
                else:
                    attrs.pop(coord)

        return attrs

    def __repr__(self):
        if self._dataset:
            return self._dataset.__repr__()
        else:
            return self.to_dataset().__repr__()

    def equals(self, other):
        if not isinstance(other, type(self)):
            return False
        # TODO: check values, etc.
        return True

    def copy(self):
        """Create a deepcopy."""
        return copy.deepcopy(self)

    @property
    def attrs(self):
        return self._attrs

    @property
    def node_dimension(self):
        """Name of node dimension"""
        return self._attrs["node_dimension"]

    @property
    def edge_dimension(self):
        """Name of edge dimension"""
        return self._attrs["edge_dimension"]

    @property
    def node_coordinates(self) -> FloatArray:
        """Coordinates (x, y) of the nodes (vertices)"""
        return np.column_stack([self.node_x, self.node_y])

    @property
    def n_node(self) -> int:
        """Number of nodes (vertices) in the UGRID topology"""
        return self.node_x.size

    @property
    def n_edge(self) -> int:
        """Number of edges in the UGRID topology"""
        return self.edge_node_connectivity.shape[0]

    @property
    def edge_x(self):
        """x-coordinate of every edge in the UGRID topology"""
        if self._edge_x is None:
            self._edge_x = self.node_x[self.edge_node_connectivity].mean(axis=1)
        return self._edge_x

    @property
    def edge_y(self):
        """y-coordinate of every edge in the UGRID topology"""
        if self._edge_y is None:
            self._edge_y = self.node_y[self.edge_node_connectivity].mean(axis=1)
        return self._edge_y

    @property
    def edge_coordinates(self) -> FloatArray:
        """Centroid (x,y) coordinates of every edge in the UGRID topology"""
        return np.column_stack([self.edge_x, self.edge_y])

    @property
    def edge_node_coordinates(self) -> FloatArray:
        """Node coordinates for every edge, shape: ``n_edge, 2, 2``."""
        return self.node_coordinates[self.edge_node_connectivity]

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        """Returns a tuple with the node bounds: xmin, ymin, xmax, ymax"""
        if any(
            [
                self._xmin is None,
                self._ymin is None,
                self._xmax is None,
                self._ymax is None,
            ]
        ):
            self._xmin = self.node_x.min()
            self._ymin = self.node_y.min()
            self._xmax = self.node_x.max()
            self._ymax = self.node_y.max()
        return (
            self._xmin,
            self._ymin,
            self._xmax,
            self._ymax,
        )

    @property
    def edge_bounds(self) -> FloatArray:
        """
        Returns a numpy array with columns ``minx, miny, maxx, maxy``,
        describing the bounds of every edge in the grid.

        Returns
        -------
        edge_bounds: np.ndarray of shape (n_edge, 4)
        """
        x = self.node_x[self.edge_node_connectivity]
        y = self.node_y[self.edge_node_connectivity]
        return np.column_stack(
            [
                x.min(axis=1),
                y.min(axis=1),
                x.max(axis=1),
                y.max(axis=1),
            ]
        )

    @staticmethod
    def _prepare_connectivity(
        da: xr.DataArray, fill_value: Union[float, int], dtype: type
    ) -> xr.DataArray:
        start_index = da.attrs.get("start_index", 0)
        if start_index not in (0, 1):
            raise ValueError(f"start_index should be 0 or 1, received: {start_index}")

        data = da.to_numpy()
        # If xarray detects a _FillValue, it converts the array to floats and
        # replaces the fill value by NaN, and moves the _FillValue to
        # da.encoding.
        if "_FillValue" in da.attrs:
            is_fill = data == da.attrs["_FillValue"]
        else:
            is_fill = np.isnan(data)
        data[is_fill] = fill_value

        cast = data.astype(dtype, copy=True)
        if start_index:
            cast -= start_index
        cast[is_fill] = fill_value
        if (cast[~is_fill] < 0).any():
            raise ValueError("connectivity contains negative values")
        return da.copy(data=cast)

    def _precheck(self, multi_index):
        dim, index = multi_index.popitem()
        for check_dim, check_index in multi_index.items():
            if not index.equals(check_index):
                raise ValueError(
                    f"UGRID dimensions do not align: {dim} versus {check_dim}"
                )
        return index

    def _postcheck(self, indexers, finalized_indexers):
        for dim, indexer in indexers.items():
            if dim != self.core_dimension:
                if not indexer.equals(finalized_indexers[dim]):
                    raise ValueError(
                        f"This subset selection of UGRID dimension {dim} results "
                        "in an invalid topology "
                    )
        return

    def set_node_coords(
        self,
        node_x: str,
        node_y: str,
        obj: Union[xr.DataArray, xr.Dataset],
        projected: bool = True,
    ):
        """
        Given names of x and y coordinates of the nodes of an object, set them
        as the coordinates in the grid.

        Parameters
        ----------
        node_x: str
            Name of the x coordinate of the nodes in the object.
        node_y: str
            Name of the y coordinate of the nodes in the object.
        """
        if " " in node_x or " " in node_y:
            raise ValueError("coordinate names may not contain spaces")

        x = obj[node_x].to_numpy()
        y = obj[node_y].to_numpy()

        if (x.ndim != 1) or (x.size != self.n_node):
            raise ValueError(
                "shape of node_x does not match n_node of grid: "
                f"{x.shape} versus {self.n_node}"
            )
        if (y.ndim != 1) or (y.size != self.n_node):
            raise ValueError(
                "shape of node_y does not match n_node of grid: "
                f"{y.shape} versus {self.n_node}"
            )

        # Remove them, then append at the end.
        node_coords = [
            coord
            for coord in self._attrs["node_coordinates"].split(" ")
            if coord not in (node_x, node_y)
        ]
        node_coords.extend((node_x, node_y))

        self._clear_geometry_properties()
        self.node_x = np.ascontiguousarray(x)
        self.node_y = np.ascontiguousarray(y)
        self._attrs["node_coordinates"] = " ".join(node_coords)
        self._indexes["node_x"] = node_x
        self._indexes["node_y"] = node_y
        self.projected = projected

    def assign_node_coords(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset

        Returns
        -------
        assigned (same type as obj)
        """
        xname = self._indexes["node_x"]
        yname = self._indexes["node_y"]
        x_attrs = conventions.DEFAULT_ATTRS["node_x"][self.projected]
        y_attrs = conventions.DEFAULT_ATTRS["node_y"][self.projected]
        coords = {
            xname: xr.DataArray(
                data=self.node_x,
                dims=(self.node_dimension,),
                attrs=x_attrs,
            ),
            yname: xr.DataArray(
                data=self.node_y,
                dims=(self.node_dimension,),
                attrs=y_attrs,
            ),
        }
        return obj.assign_coords(coords)

    def assign_edge_coords(
        self,
        obj: Union[xr.DataArray, xr.Dataset],
    ) -> Union[xr.DataArray, xr.Dataset]:
        """
        Assign node coordinates from the grid to the object.

        Returns a new object with all the original data in addition to the new
        node coordinates of the grid.

        Parameters
        ----------
        obj: xr.DataArray or xr.Dataset

        Returns
        -------
        assigned (same type as obj)
        """
        xname = self._indexes.get("edge_x", f"{self.name}_edge_x")
        yname = self._indexes.get("edge_y", f"{self.name}_edge_y")
        x_attrs = conventions.DEFAULT_ATTRS["edge_x"][self.projected]
        y_attrs = conventions.DEFAULT_ATTRS["edge_y"][self.projected]
        coords = {
            xname: xr.DataArray(
                data=self.edge_x,
                dims=(self.edge_dimension,),
                attrs=x_attrs,
            ),
            yname: xr.DataArray(
                data=self.edge_y,
                dims=(self.edge_dimension,),
                attrs=y_attrs,
            ),
        }
        return obj.assign_coords(coords)

    @property
    def node_edge_connectivity(self) -> csr_matrix:
        """
        Node to edge connectivity.

        Returns
        -------
        connectivity: csr_matrix
        """
        if self._node_edge_connectivity is None:
            self._node_edge_connectivity = connectivity.invert_dense_to_sparse(
                self.edge_node_connectivity, self.fill_value
            )
        return self._node_edge_connectivity

    @property
    def node_node_connectivity(self) -> csr_matrix:
        """
        Node to node connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) node index. The data of
        the matrix contains the edge index as every connection is formed by an
        edge.

        Returns
        -------
        connectivity: csr_matrix
        """
        if self._node_node_connectivity is None:
            self._node_node_connectivity = connectivity.node_node_connectivity(
                self.edge_node_connectivity
            )
        return self._node_node_connectivity

    @property
    def directed_node_node_connectivity(self) -> csr_matrix:
        """
        Directed node to node connectivity.

        The connectivity is represented as an adjacency matrix in CSR format,
        with the row and column indices as a (0-based) node index. The data of
        the matrix contains the edge index as every connection is formed by an
        edge.

        Returns
        -------
        connectivity: csr_matrix
        """
        return connectivity.directed_node_node_connectivity(self.edge_node_connectivity)

    @staticmethod
    def _connectivity_weights(connectivity: csr_matrix, coordinates: FloatArray):
        xy = coordinates
        coo = connectivity.tocoo()
        i = coo.row
        j = coo.col
        distance = np.linalg.norm(xy[j] - xy[i], axis=1)
        # The inverse distance is a measure of the strength of the connection.
        # Normalize so the weights are around 1.0
        return distance.mean() / distance

    def set_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
        allow_override: bool = False,
    ):
        """
        Set the Coordinate Reference System (CRS) of a UGRID topology.

        NOTE: The underlying geometries are not transformed to this CRS. To
        transform the geometries to a new CRS, use the ``to_crs`` method.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted
            by :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying the projection.
        allow_override : bool, default False
            If the the UGRID topology already has a CRS, allow to replace the
            existing CRS, even when both are not equal.
        """
        import pyproj

        if crs is not None:
            crs = pyproj.CRS.from_user_input(crs)
        elif epsg is not None:
            crs = pyproj.CRS.from_epsg(epsg)
        else:
            raise ValueError("Must pass either crs or epsg.")

        if not allow_override and self.crs is not None and not self.crs == crs:
            raise ValueError(
                "The Ugrid already has a CRS which is not equal to the passed "
                "CRS. Specify 'allow_override=True' to allow replacing the existing "
                "CRS without doing any transformation. If you actually want to "
                "transform the geometries, use '.to_crs' instead."
            )
        self.crs = crs

    def to_crs(
        self,
        crs: Union["pyproj.CRS", str] = None,  # type: ignore # noqa
        epsg: int = None,
    ):
        """
        Transform geometries to a new coordinate reference system.
        Transform all geometries in an active geometry column to a different coordinate
        reference system. The ``crs`` attribute on the current Ugrid must
        be set. Either ``crs`` or ``epsg`` may be specified for output.

        This method will transform all points in all objects. It has no notion
        of projecting the cells. All segments joining points are assumed to be
        lines in the current projection, not geodesics. Objects crossing the
        dateline (or other projection boundary) will have undesirable behavior.

        Parameters
        ----------
        crs : pyproj.CRS, optional if `epsg` is specified
            The value can be anything accepted by
            :meth:`pyproj.CRS.from_user_input() <pyproj.crs.CRS.from_user_input>`,
            such as an authority string (eg "EPSG:4326") or a WKT string.
        epsg : int, optional if `crs` is specified
            EPSG code specifying output projection.
        """
        import pyproj

        if self.crs is None:
            raise ValueError(
                "Cannot transform naive geometries.  "
                "Please set a crs on the object first."
            )
        if crs is not None:
            crs = pyproj.CRS.from_user_input(crs)
        elif epsg is not None:
            crs = pyproj.CRS.from_epsg(epsg)
        else:
            raise ValueError("Must pass either crs or epsg.")

        grid = self.copy()
        if self.crs.is_exact_same(crs):
            return grid

        transformer = pyproj.Transformer.from_crs(
            crs_from=self.crs, crs_to=crs, always_xy=True
        )
        node_x, node_y = transformer.transform(xx=grid.node_x, yy=grid.node_y)
        grid.node_x = node_x
        grid.node_y = node_y
        grid._clear_geometry_properties()
        grid._dataset = None
        grid.crs = crs

        return grid

    @property
    def is_geographic(self):
        if self.crs is None:
            return False
        return self.crs.is_geographic

    def plot(self, **kwargs):
        """
        Plot the edges of the mesh.

        Parameters
        ----------
        **kwargs : optional
            Additional keyword arguments to ``matplotlib.pyplot.line``.
        """
        from xugrid.plot import line

        return line(self, **kwargs)


UgridType = Type[AbstractUgrid]



import numpy as np
import pytest
import xarray as xr

import xugrid as xu
from xugrid import (
    BarycentricInterpolator,
    CentroidLocatorRegridder,
    OverlapRegridder,
    RelativeOverlapRegridder,
)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_check_source_target_types(disk, cls):
    with pytest.raises(TypeError):
        cls(source=disk, target=1.0)
    with pytest.raises(TypeError):
        cls(source=1.0, target=disk)


@pytest.mark.parametrize(
    "regridder_class",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_structured_to_unstructured(
    regridder_class,
    disk,
    quads_structured,
):
    regridder = regridder_class(quads_structured, disk)
    actual = regridder.regrid(quads_structured)
    assert isinstance(actual, xu.UgridDataArray)


def test_centroid_locator_regridder_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_centroid
):
    regridder = CentroidLocatorRegridder(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result.fillna(0.0) == expected_results_centroid.fillna(0.0)).any()

    # With broadcasting
    regridder = CentroidLocatorRegridder(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (
        broadcasted.fillna(0.0).isel(layer=0) == expected_results_centroid.fillna(0.0)
    ).any()


def test_centroid_locator_regridder(disk, quads_1):
    square = quads_1
    regridder = CentroidLocatorRegridder(source=disk, target=square)
    result = regridder.regrid(disk)
    assert isinstance(result, xu.UgridDataArray)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()
    assert result.grid.n_face == square.grid.n_face

    # other way around
    regridder = CentroidLocatorRegridder(source=result, target=disk)
    back = regridder.regrid(result)
    assert isinstance(back, xu.UgridDataArray)
    assert back.notnull().any()
    assert back.min() >= disk.min()
    assert back.max() <= disk.max()
    assert back.grid.n_face == disk.grid.n_face

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * result.obj,
        result.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", disk.grid.face_dimension)


def test_overlap_regridder_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_overlap
):
    regridder = OverlapRegridder(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result == expected_results_overlap).any()

    # With broadcasting
    regridder = OverlapRegridder(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (broadcasted.isel(layer=0) == expected_results_overlap).any()


def test_overlap_regridder(disk, quads_1):
    square = quads_1
    regridder = OverlapRegridder(disk, square, method="mean")
    result = regridder.regrid(disk)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * disk.obj,
        grid=disk.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", square.grid.face_dimension)
    assert broadcasted.shape == (5, 100)


def test_linear_interpolator_structured(
    grid_data_a, grid_data_a_layered, grid_data_b, expected_results_linear
):
    regridder = BarycentricInterpolator(source=grid_data_a, target=grid_data_b)
    result = regridder.regrid(grid_data_a)
    assert (result.fillna(0.0) == expected_results_linear.fillna(0.0)).any()

    # With broadcasting
    regridder = BarycentricInterpolator(source=grid_data_a_layered, target=grid_data_b)
    broadcasted = regridder.regrid(grid_data_a_layered)
    assert broadcasted.dims == ("layer", "y", "x")
    assert (
        broadcasted.fillna(0.0).isel(layer=0) == expected_results_linear.fillna(0.0)
    ).any()


def test_barycentric_interpolator(disk, quads_0_25):
    square = quads_0_25
    regridder = BarycentricInterpolator(source=disk, target=square)
    result = regridder.regrid(disk)
    assert result.notnull().any()
    assert result.min() >= disk.min()
    assert result.max() <= disk.max()

    # With broadcasting
    obj = xu.UgridDataArray(
        xr.DataArray(np.ones(5), dims=["layer"]) * disk.obj,
        grid=disk.grid,
    )
    broadcasted = regridder.regrid(obj)
    assert broadcasted.dims == ("layer", square.grid.face_dimension)
    assert broadcasted.shape == (5, 1600)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_weights(cls, disk, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    weights = regridder.weights
    new_regridder = cls.from_weights(weights, target=square)
    new_result = new_regridder.regrid(disk)
    assert new_result.equals(result)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_weights_layered(cls, disk, disk_layered, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    weights = regridder.weights
    new_regridder = cls.from_weights(weights, target=square)
    new_result = new_regridder.regrid(disk_layered)
    assert np.array_equal(new_result.sel(layer=1).values, result.values, equal_nan=True)


@pytest.mark.parametrize(
    "cls",
    [
        CentroidLocatorRegridder,
        OverlapRegridder,
        RelativeOverlapRegridder,
        BarycentricInterpolator,
    ],
)
def test_regridder_from_dataset(cls, disk, quads_1):
    square = quads_1
    regridder = cls(source=disk, target=square)
    result = regridder.regrid(disk)
    dataset = regridder.to_dataset()
    new_regridder = cls.from_dataset(dataset)
    new_result = new_regridder.regrid(disk)
    assert np.array_equal(new_result.values, result.values, equal_nan=True)


def test_regridder_daks_arrays(
    grid_data_dask_source,
    grid_data_dask_source_layered,
    grid_data_dask_target,
    grid_data_dask_expected,
    grid_data_dask_expected_layered,
):
    regridder = CentroidLocatorRegridder(
        source=grid_data_dask_source, target=grid_data_dask_target
    )
    result = regridder.regrid(grid_data_dask_source)
    assert result.equals(grid_data_dask_expected)

    # with broadcasting
    regridder = CentroidLocatorRegridder(
        source=grid_data_dask_source_layered, target=grid_data_dask_target
    )
    result = regridder.regrid(grid_data_dask_source_layered)
    assert result.isel(layer=0).equals(grid_data_dask_expected_layered.isel(layer=0))



from __future__ import annotations

from typing import NamedTuple, Tuple

import numba as nb
import numpy as np
import pandas as pd
from scipy import sparse

from xugrid.constants import BoolArray, FloatArray, IntArray, IntDType, SparseMatrix


def argsort_rows(array: np.ndarray) -> IntArray:
    if array.ndim != 2:
        raise ValueError(f"Array is not 2D, but has shape: {array.shape}")
    dtype = [("f{i}".format(i=i), array.dtype) for i in range(array.shape[1])]
    arr1d = array.view(dtype).flatten()
    return np.argsort(arr1d)


def index_like(xy_a: FloatArray, xy_b: FloatArray, tolerance: float):
    """Return the index that would transform xy_a into xy_b."""
    if xy_a.shape != xy_b.shape:
        raise ValueError("coordinates do not match in shape")

    # Quantize coordinates before sorting in case a tolerance is given.
    # After quantization, coordinates can be considered integers.
    # Float64 has ample room for our coordinate system even with small
    # tolerances, e.g. 1.0e9 / 1.0e-9 is not a problem.
    if tolerance != 0.0:
        sorter_a = argsort_rows(np.round(xy_a / tolerance))
        sorter_b = argsort_rows(np.round(xy_b / tolerance))
    else:
        sorter_a = argsort_rows(xy_a)
        sorter_b = argsort_rows(xy_b)

    if not np.allclose(xy_a[sorter_a], xy_b[sorter_b], rtol=0.0, atol=tolerance):
        raise ValueError("coordinates are not identical after sorting")
    #
    #   sorter_a         inverse_b
    # a --------> sorted ---------> b
    #
    inverse_b = np.argsort(sorter_b)
    return sorter_a[inverse_b]


class AdjacencyMatrix(NamedTuple):
    indices: IntArray
    indptr: IntArray
    nnz: int
    n: int
    m: int


def _csr_to_adjacency(A: sparse.csr_matrix) -> AdjacencyMatrix:
    if not isinstance(A, sparse.csr_matrix):
        raise TypeError(
            f"Expected scipy.sparse.csr_matrix, received: {type(A).__name__}"
        )

    n, m = A.shape
    adj = AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)
    return adj


@nb.njit(inline="always")
def neighbors(A: AdjacencyMatrix, cell: int) -> IntArray:
    start = A.indptr[cell]
    end = A.indptr[cell + 1]
    return A.indices[start:end]


@nb.njit(inline="always")
def pop(array, size):
    return array[size - 1], size - 1


@nb.njit(inline="always")
def push(array, value, size):
    array[size] = value
    return size + 1


@nb.njit
def _topological_sort_by_dfs(A: AdjacencyMatrix):
    # This code is almost a direct port of the BSD-2 licensed code in Graphs.jl:
    #
    # https://github.com/JuliaGraphs/Graphs.jl/blob/master/LICENSE.md
    #
    # Copyright (c) 2015: Seth Bromberger and other contributors. Copyright (c)
    # 2012: John Myles White and other contributors.
    #
    # Redistribution and use in source and binary forms, with or without
    # modification, are permitted provided that the following conditions are met:
    #
    # Redistributions of source code must retain the above copyright notice, this
    # list of conditions and the following disclaimer. Redistributions in binary
    # form must reproduce the above copyright notice, this list of conditions and
    # the following disclaimer in the documentation and/or other materials provided
    # with the distribution. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
    # CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
    # NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    # PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    # CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    # EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    # PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
    # OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
    # WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
    # OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
    # ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    vcolor = np.zeros(A.m, dtype=np.uint8)
    verts = np.empty(A.m, dtype=np.int64)
    verts_size = 0
    S = np.empty(A.m, dtype=np.int64)
    S_size = 0

    for v in range(A.m):
        if vcolor[v] != 0:
            continue
        S_size = 0
        S_size = push(S, v, S_size)
        vcolor[v] = 1
        while S_size > 0:
            u = S[S_size - 1]
            w = 0
            for n in neighbors(A, u):
                if vcolor[n] == 1:
                    raise ValueError("The graph contains at least one cycle")
                elif vcolor[n] == 0:
                    w = n
                    break

            if w != 0:
                vcolor[w] = 1
                S_size = push(S, w, S_size)
            else:
                vcolor[u] = 2
                verts_size = push(verts, u, verts_size)
                S_size -= 1

    return verts[::-1]


def topological_sort_by_dfs(A: sparse.csr_matrix) -> IntArray:
    """
    Return an array of vertices in topological order.

    Parameters
    ----------
    A: sparse.csr_matrix
        CSR adjacency matrix of the directed acyclical graph.

    Returns
    -------
    sorted_vertices: np.ndarray of integer
    """
    return _topological_sort_by_dfs(_csr_to_adjacency(A))


@nb.njit
def _contract_vertices(A: AdjacencyMatrix, indices: IntArray) -> IntArray:
    vcolor = np.zeros(A.m, dtype=np.uint8)
    vcolor[indices] = 2

    edge_node_connectivity = np.empty((A.m, 2), dtype=np.int64)
    n_edge = 0

    S = np.empty(A.m, dtype=np.int64)
    size = 0

    for v in indices:
        size = 0
        size = push(S, v, size)
        while size > 0:
            u, size = pop(S, size)

            # Do not paint over the marked indices
            if vcolor[u] == 0:
                vcolor[u] == 1

            for n in neighbors(A, u):
                if (n == v) or (vcolor[n] == 1):
                    raise ValueError("The graph contains at least one cycle")
                elif vcolor[n] == 2:
                    edge_node_connectivity[n_edge, 0] = v
                    edge_node_connectivity[n_edge, 1] = n
                    n_edge += 1
                    size -= 1
                else:
                    size = push(S, n, size)

    return edge_node_connectivity[:n_edge]


def contract_vertices(A: sparse.csr_matrix, indices: IntArray) -> IntArray:
    """
    Contract vertices to the set defined by indices.

    Parameters
    ----------
    A: AdjacencyMatrix
        CSR adjacency matrix of the directed acyclical graph.
    indices: np.ndarray of integer
        Which vertices to preserve.

    Returns
    -------
    edge_node_connectivity: np.ndarray of shape (n_edge, 2)
        New edge_node_connectivity containing edges of the vertices contained
        in indices.
    """
    return _contract_vertices(_csr_to_adjacency(A), np.array(indices))


# Conversion between dense and sparse
# -----------------------------------
def _to_ij(conn: IntArray, fill_value: int, invert: bool) -> Tuple[IntArray, IntArray]:
    n, m = conn.shape
    j = conn.ravel()
    valid = j != fill_value
    i = np.repeat(np.arange(n), m)[valid]
    j = j[valid]
    if invert:
        return j, i
    else:
        return i, j


def _to_sparse(
    conn: IntArray, fill_value: int, invert: bool, sort_indices: bool
) -> sparse.csr_matrix:
    i, j = _to_ij(conn, fill_value, invert)
    coo_content = (j, (i, j))
    coo_matrix = sparse.coo_matrix(coo_content)
    csr_matrix = coo_matrix.tocsr()
    # Conversion to csr format results in a sorting of indices. We require
    # only sorting of i, not j, as this would e.g. mess up the order for
    # counter clockwise vertex orientation, e.g.
    if not sort_indices:
        order = np.argsort(i)
        csr_matrix.indices = j[order]
        csr_matrix.has_sorted_indices = False
    return csr_matrix


def ragged_index(n: int, m: int, m_per_row: IntArray) -> BoolArray:
    """
    Given an array of n rows by m columns, starting from left mark the values
    True such that the number of True values equals m_per_row.

    For example:

    n = 3
    m = 4
    m_per_row = np.array([1, 2, 3])

    Then the result of _ragged_index(n, m, m_per_row) is:

    np.array([
        [True, False, False, False],
        [True, True, False, False],
        [True, True, True, False],
    ])

    This can be used as boolean index to set a variable number of values per
    row.
    """
    column_number = np.tile(np.arange(m), n).reshape((n, m))
    return (column_number.T < m_per_row).T


def to_sparse(
    conn: IntArray, fill_value: int, sort_indices: bool = True
) -> sparse.csr_matrix:
    return _to_sparse(conn, fill_value, invert=False, sort_indices=sort_indices)


def to_dense(conn: SparseMatrix, fill_value: int, n_columns: int = None) -> IntArray:
    n, _ = conn.shape
    m_per_row = conn.getnnz(axis=1)
    m = m_per_row.max()
    if n_columns is not None:
        if n_columns < m:
            raise ValueError(
                f"n_columns {n_columns} is too small for the data, requires {m}"
            )
        m = n_columns

    # Allocate 2D array and create a flat view of the dense connectivity
    dense_conn = np.empty((n, m), dtype=IntDType)
    flat_conn = dense_conn.ravel()
    if (n * m) == conn.nnz:
        # Shortcut if fill_value is not present, when all of same geom. type
        # e.g. all triangles or all quadrangles
        valid = slice(None)  # a[:] equals a[slice(None)]
    else:
        valid = ragged_index(n, m, m_per_row).ravel()
        flat_conn[~valid] = fill_value

    if isinstance(conn, sparse.csr_matrix):
        flat_conn[valid] = conn.indices
    elif isinstance(conn, sparse.coo_matrix):
        flat_conn[valid] = conn.col
    else:
        raise TypeError("Can only invert coo or csr matrix")
    return dense_conn


# Inverting connectivities
# ------------------------
def invert_dense_to_sparse(
    conn: IntArray, fill_value: int, sort_indices: bool = True
) -> sparse.csr_matrix:
    return _to_sparse(conn, fill_value, invert=True, sort_indices=sort_indices)


def invert_dense(
    conn: IntArray, fill_value: int, sort_indices: bool = True
) -> IntArray:
    sparse_inverted = _to_sparse(
        conn, fill_value, invert=True, sort_indices=sort_indices
    )
    return to_dense(sparse_inverted, fill_value)


def invert_sparse(conn: sparse.csr_matrix) -> sparse.csr_matrix:
    coo = conn.tocoo()
    j = coo.row
    i = coo.col
    coo_content = (j, (i, j))
    inverted = sparse.coo_matrix(coo_content)
    return inverted.tocsr()


def invert_sparse_to_dense(conn: sparse.csr_matrix, fill_value: int) -> IntArray:
    inverted = invert_sparse(conn)
    return to_dense(inverted, fill_value)


# Renumbering
# -----------
def _renumber(a: IntArray) -> IntArray:
    # Taken from https://github.com/scipy/scipy/blob/v1.7.1/scipy/stats/stats.py#L8631-L8737
    # (scipy is BSD-3-Clause License)
    arr = np.ravel(np.asarray(a))
    sorter = np.argsort(arr, kind="quicksort")
    inv = np.empty(sorter.size, dtype=IntDType)
    inv[sorter] = np.arange(sorter.size, dtype=IntDType)
    arr = arr[sorter]
    obs = np.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv] - 1
    return dense.reshape(a.shape)


def renumber(a: IntArray, fill_value: int = None):
    if fill_value is None:
        return _renumber(a)

    valid = a != fill_value
    renumbered = np.full_like(a, fill_value)
    renumbered[valid] = _renumber(a[valid])
    return renumbered


def close_polygons(face_node_connectivity: IntArray, fill_value: int) -> IntArray:
    # Wrap around and create closed polygon: put the first node at the end of the row
    # In case of fill values, replace all fill values
    n, m = face_node_connectivity.shape
    closed = np.full((n, m + 1), fill_value, dtype=IntDType)
    closed[:, :-1] = face_node_connectivity
    first_node = face_node_connectivity[:, 0]
    # Identify fill value, and replace by first node also
    isfill = closed == fill_value
    closed.ravel()[isfill.ravel()] = np.repeat(first_node, isfill.sum(axis=1))
    return closed, isfill


def reverse_orientation(face_node_connectivity: IntArray, fill_value: int):
    # We cannot simply reverse the rows with [:, ::-1], since there may be fill
    # values present.
    reversed_orientation = face_node_connectivity.copy()
    in_reverse = face_node_connectivity[:, ::-1]
    in_reverse = in_reverse[in_reverse != fill_value]
    replace = face_node_connectivity != fill_value
    reversed_orientation[replace] = in_reverse
    return reversed_orientation


def counterclockwise(
    face_node_connectivity: IntArray, fill_value: int, nodes: FloatArray
) -> IntArray:
    # TODO: in case of "periodic grids", we need to wrap around the grid.
    closed, _ = close_polygons(face_node_connectivity, fill_value)
    p = nodes[closed]
    dxy = np.diff(p, axis=1)
    reverse = (np.cross(dxy[:, :-1], dxy[:, 1:])).sum(axis=1) < 0
    ccw = face_node_connectivity.copy()
    if reverse.any():
        ccw[reverse] = reverse_orientation(face_node_connectivity[reverse], fill_value)
    return ccw


# Derived connectivities
# ----------------------
def boundary_node_connectivity(
    edge_face_connectivity: IntArray,
    fill_value: int,
    edge_node_connectivity: IntArray,
) -> IntArray:
    """Is a subset of the edge_node_connectivity"""
    is_boundary = (edge_face_connectivity == fill_value).any(axis=1)
    return edge_node_connectivity[is_boundary]


def edge_connectivity(
    face_node_connectivity: IntArray,
    fill_value: int,
    edge_node_connectivity=None,
) -> Tuple[IntArray, IntArray]:
    """Derive new edge_node_connectivity and face_edge_connectivity."""
    prior = edge_node_connectivity
    n, m = face_node_connectivity.shape
    # Close the polygons: [0 1 2 3] -> [0 1 2 3 0]
    closed, isfill = close_polygons(face_node_connectivity, fill_value)
    # Allocate array for edge_node_connectivity: includes duplicate edges
    edge_node_connectivity = np.empty((n * m, 2), dtype=IntDType)
    edge_node_connectivity[:, 0] = closed[:, :-1].ravel()
    edge_node_connectivity[:, 1] = closed[:, 1:].ravel()
    # Cleanup: delete invalid edges (same node to same node)
    edge_node_connectivity = edge_node_connectivity[
        edge_node_connectivity[:, 0] != edge_node_connectivity[:, 1]
    ]
    # Now find the unique rows == unique edges
    edge_node_connectivity.sort(axis=1)
    edge_node_connectivity, inverse_indices = np.unique(
        ar=edge_node_connectivity, return_inverse=True, axis=0
    )

    if prior is not None:  # prior edge_node_connectivity exists
        unique, index = np.unique(np.sort(prior, axis=1), axis=0, return_index=True)
        # Check whether everything looks okay:
        if not np.array_equal(unique, edge_node_connectivity):
            raise ValueError(
                "Invalid edge_node_connectivity. Run .validate_edge_node_connectivity()."
            )
        inverse_indices = index[inverse_indices]
        edge_node_connectivity = prior

    # Create face_edge_connectivity
    face_edge_connectivity = np.full((n, m), fill_value, dtype=np.int64)
    isnode = ~isfill[:, :-1]
    face_edge_connectivity[isnode] = inverse_indices
    return edge_node_connectivity, face_edge_connectivity


def validate_edge_node_connectivity(
    face_node_connectivity: IntArray,
    fill_value: int,
    edge_node_connectivity: IntArray,
) -> BoolArray:
    """
    * Is the edge defined by the face_node_connectivity?
    * Are the edges unique?
    """
    new, _ = edge_connectivity(face_node_connectivity, fill_value)
    old = np.sort(edge_node_connectivity, axis=1)

    new1d = new.astype(np.int32).view(np.int64).ravel()
    old1d = old.astype(np.int32).view(np.int64).ravel()

    s = pd.Series(old1d)
    n_edge = len(new1d)
    n_old = s.nunique()
    if n_old < n_edge:
        raise ValueError(
            f"face_node_connectivity defines {n_edge} edges, but "
            f"edge_node_connectivity defines only {n_old} edges."
        )

    valid = np.isin(old1d, new1d) & (~s.duplicated().to_numpy())
    return valid


def face_face_connectivity(
    edge_face_connectivity: IntArray,
    fill_value: int,
) -> sparse.csr_matrix:
    """
    Derive face to face connectivity.

    An edge can be shared by two faces at most. If this is the case, they are
    neighbors.
    """
    i = edge_face_connectivity[:, 0]
    j = edge_face_connectivity[:, 1]
    is_connection = j != fill_value
    i = i[is_connection]
    j = j[is_connection]
    edge_index = np.arange(len(edge_face_connectivity))[is_connection]
    ij = np.concatenate([i, j])
    ji = np.concatenate([j, i])
    edge_index = np.concatenate([edge_index, edge_index])
    coo_content = (edge_index, (ij, ji))
    coo_matrix = sparse.coo_matrix(coo_content)
    return coo_matrix.tocsr()


def directed_node_node_connectivity(
    edge_node_connectivity: IntArray,
) -> sparse.csr_matrix:
    i = edge_node_connectivity[:, 0]
    j = edge_node_connectivity[:, 1]
    edge_index = np.arange(len(edge_node_connectivity))
    coo_content = (edge_index, (i, j))
    n = max(i.max(), j.max()) + 1
    coo_matrix = sparse.coo_matrix(coo_content, shape=(n, n))
    return coo_matrix.tocsr()


def node_node_connectivity(edge_node_connectivity: IntArray) -> sparse.csr_matrix:
    i = edge_node_connectivity[:, 0]
    j = edge_node_connectivity[:, 1]
    edge_index = np.arange(len(edge_node_connectivity))
    ij = np.concatenate([i, j])
    ji = np.concatenate([j, i])
    edge_index = np.concatenate([edge_index, edge_index])
    coo_content = (edge_index, (ij, ji))
    coo_matrix = sparse.coo_matrix(coo_content)
    return coo_matrix.tocsr()


def structured_connectivity(active: IntArray) -> AdjacencyMatrix:
    nrow, ncol = active.shape
    nodes = np.arange(nrow * ncol).reshape(nrow, ncol)
    nodes[~active] = -1
    left = nodes[:, :-1].ravel()
    right = nodes[:, 1:].ravel()
    front = nodes[:-1].ravel()
    back = nodes[1:].ravel()
    # valid x connection
    valid = (left != -1) & (right != -1)
    left = left[valid]
    right = right[valid]
    # valid y connection
    valid = (front != -1) & (back != -1)
    front = front[valid]
    back = back[valid]
    i = renumber(np.concatenate([left, right, front, back]))
    j = renumber(np.concatenate([right, left, back, front]))
    coo_content = (j, (i, j))
    A = sparse.coo_matrix(coo_content).tocsr()
    n, m = A.shape
    return AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)


def perimeter(
    face_node_connectivity: IntArray,
    fill_value: int,
    node_x: FloatArray,
    node_y: FloatArray,
):
    nodes = np.column_stack([node_x, node_y])
    closed, _ = close_polygons(face_node_connectivity, fill_value)
    coordinates = nodes[closed]
    # Shift coordinates to avoid precision loss
    xy0 = coordinates[:, 0]
    coordinates -= xy0[:, np.newaxis]
    dxy = np.diff(coordinates, axis=1)
    return np.linalg.norm(dxy, axis=-1).sum(axis=1)


def area(
    face_node_connectivity: IntArray,
    fill_value: int,
    node_x: FloatArray,
    node_y: FloatArray,
):
    nodes = np.column_stack([node_x, node_y])
    closed, _ = close_polygons(face_node_connectivity, fill_value)
    coordinates = nodes[closed]
    xy0 = coordinates[:, 0]
    a = coordinates[:, :-1] - xy0[:, np.newaxis]
    b = coordinates[:, 1:] - xy0[:, np.newaxis]
    determinant = np.cross(a, b)
    return 0.5 * abs(determinant.sum(axis=1))


def centroids(
    face_node_connectivity: IntArray,
    fill_value: int,
    node_x: FloatArray,
    node_y: FloatArray,
) -> FloatArray:
    n_face, n_max_node = face_node_connectivity.shape
    nodes = np.column_stack([node_x, node_y])
    # Check if it's fully triangular
    if n_max_node == 3:
        # Since it's triangular, computing the average of the vertices suffices
        coordinates = nodes[face_node_connectivity]
        return np.nanmean(coordinates, axis=1)
    else:
        # This is mathematically equivalent to triangulating, computing triangle centroids
        # and computing the area weighted average of those centroids
        centroid_coordinates = np.empty((n_face, 2), dtype=np.float64)
        closed, _ = close_polygons(face_node_connectivity, fill_value)
        coordinates = nodes[closed]
        # Express coordinates relative to first node
        xy0 = coordinates[:, 0]
        a = coordinates[:, :-1] - xy0[:, np.newaxis]
        b = coordinates[:, 1:] - xy0[:, np.newaxis]
        c = a + b
        determinant = np.cross(a, b)
        area_weight = 1.0 / (3.0 * determinant.sum(axis=1))
        centroid_coordinates[:, 0] = area_weight * (c[..., 0] * determinant).sum(axis=1)
        centroid_coordinates[:, 1] = area_weight * (c[..., 1] * determinant).sum(axis=1)
        centroid_coordinates += xy0
        return centroid_coordinates


@nb.njit(cache=True, parallel=True)
def _circumcenters_triangle(xxx: FloatArray, yyy: FloatArray):
    """Numba should nicely fuse these operations."""
    a_x, b_x, c_x = xxx
    a_y, b_y, c_y = yyy
    D_inv = 0.5 / (
        a_y * c_x + b_y * a_x - b_y * c_x - a_y * b_x - c_y * a_x + c_y * b_x
    )
    x = ((a_x - c_x) * (a_x + c_x) + (a_y - c_y) * (a_y + c_y)) * (b_y - c_y) - (
        (b_x - c_x) * (b_x + c_x) + (b_y - c_y) * (b_y + c_y)
    ) * (a_y - c_y)
    y = ((b_x - c_x) * (b_x + c_x) + (b_y - c_y) * (b_y + c_y)) * (a_x - c_x) - (
        (a_x - c_x) * (a_x + c_x) + (a_y - c_y) * (a_y + c_y)
    ) * (b_x - c_x)
    return D_inv * x, D_inv * y


def circumcenters(
    face_node_connectivity: IntArray,
    fill_value: int,
    node_x: FloatArray,
    node_y: FloatArray,
) -> FloatArray:
    # TODO: Skyum or Welzl implementation for polygons -- although it's
    # practical use is dubious?
    n_max_node = face_node_connectivity.shape[1]
    # Check if it's fully triangular
    if n_max_node == 3:
        xxx = node_x[face_node_connectivity.T]
        yyy = node_y[face_node_connectivity.T]
        x, y = _circumcenters_triangle(xxx, yyy)
    else:
        raise NotImplementedError(
            "Circumcenters are only supported for triangular grids"
        )
    return np.column_stack((x, y))


def _triangulate(i: IntArray, j: IntArray, n_triangle_per_row: IntArray) -> IntArray:
    n_triangle = n_triangle_per_row.sum()
    n_face = len(i)
    index_first = np.argwhere(np.diff(i, prepend=-1) != 0)
    index_second = index_first + 1
    index_last = np.argwhere(np.diff(i, append=-1) != 0)

    first = np.full(n_face, False)
    first[index_first] = True
    second = np.full(n_face, True) & ~first
    second[index_last] = False
    third = np.full(n_face, True) & ~first
    third[index_second] = False

    triangles = np.empty((n_triangle, 3), IntDType)
    triangles[:, 0] = np.repeat(j[first], n_triangle_per_row)
    triangles[:, 1] = j[second]
    triangles[:, 2] = j[third]
    return triangles


def triangulate_dense(
    face_node_connectivity: IntArray, fill_value: int
) -> Tuple[IntArray, IntArray]:
    n_face, n_max = face_node_connectivity.shape

    if n_max == 3:
        triangles = face_node_connectivity.copy()
        return triangles, np.arange(n_face)

    valid = face_node_connectivity != fill_value
    n_per_row = valid.sum(axis=1)
    n_triangle_per_row = n_per_row - 2
    i = np.repeat(np.arange(n_face), n_per_row)
    j = face_node_connectivity.ravel()[valid.ravel()]
    triangles = _triangulate(i, j, n_triangle_per_row)

    triangle_face_connectivity = np.repeat(
        np.arange(n_face), repeats=n_triangle_per_row
    )
    return triangles, triangle_face_connectivity


def triangulate_coo(
    face_node_connectivity: sparse.coo_matrix
) -> Tuple[IntArray, IntArray]:
    ncol_per_row = face_node_connectivity.getnnz(axis=1)

    if ncol_per_row.max() == 3:
        triangles = face_node_connectivity.row.copy().reshape((-1, 3))
        return triangles, np.arange(len(triangles))

    n_triangle_per_row = ncol_per_row - 2
    i = face_node_connectivity.row
    j = face_node_connectivity.col
    triangles = _triangulate(i, j, n_triangle_per_row)

    n_face = face_node_connectivity.shape[0]
    triangle_face_connectivity = np.repeat(
        np.arange(n_face), repeats=n_triangle_per_row
    )
    return triangles, triangle_face_connectivity


def triangulate(
    face_node_connectivity, fill_value: int | None = None
) -> Tuple[IntArray, IntArray]:
    """
    Convert polygons into its constituent triangles.

    Triangulation runs from the first node of every face:

    * first, second, third,
    * first, third, fourth,
    * and so forth ...

    If the grid is already triangular, a copy is returned.

    Returns
    -------
    triangles: ndarray of integers with shape ``(n_triangle, 3)``
    triangle_face_connectivity: ndarray of integers with shape ``(n_triangle,)``
    """
    if isinstance(face_node_connectivity, IntArray):
        if fill_value is None:
            raise ValueError("fill_value is required for dense connectivity")
        return triangulate_dense(face_node_connectivity, fill_value)
    elif isinstance(face_node_connectivity, sparse.coo_matrix):
        return triangulate_coo(face_node_connectivity)
    else:
        raise TypeError("connectivity must be ndarray or sparse matrix")


def _mutate(
    output: BoolArray, i: IntArray, j: IntArray, value: bool, mask: BoolArray
) -> None:
    a = output[i]
    b = output[j]
    mutate = a != b
    output[i[mutate]] = value
    output[j[mutate]] = value
    if mask is not None:
        output[mask] = not value
    return None


def _binary_iterate(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    value: bool,
    iterations: int,
    mask: BoolArray,
    exterior: IntArray,
    border_value: bool,
) -> BoolArray:
    if input.dtype != np.bool_:
        raise TypeError("input dtype should be bool")

    ndim = input.ndim
    if ndim != 1:
        raise ValueError(
            "Binary operations are only supported for a single (face)"
            f"dimension. Found {ndim} dimensions."
        )

    coo = connectivity.tocoo()
    i = coo.row
    j = coo.col
    output = input.copy()

    # First iteration, mutate borders.
    _mutate(output, i, j, value, mask)
    if exterior is not None and value == border_value:
        output[exterior] = value

    # Subsequent iterations, disregard border.
    for _ in range(iterations - 1):
        _mutate(output, i, j, value, mask)

    return output


def binary_erosion(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    iterations: int = 1,
    mask: BoolArray | None = None,
    exterior: IntArray | None = None,
    border_value: bool | None = False,
) -> BoolArray:
    """By default, erodes inwards from the exterior."""
    return _binary_iterate(
        connectivity=connectivity,
        input=input,
        value=False,
        iterations=iterations,
        mask=mask,
        exterior=exterior,
        border_value=border_value,
    )


def binary_dilation(
    connectivity: sparse.csr_matrix,
    input: BoolArray,
    iterations: int = 1,
    mask: BoolArray | None = None,
    exterior: IntArray | None = None,
    border_value: bool | None = False,
) -> BoolArray:
    """By default, does not dilate inward from the exterior."""
    return _binary_iterate(
        connectivity=connectivity,
        input=input,
        value=True,
        iterations=iterations,
        mask=mask,
        exterior=exterior,
        border_value=border_value,
    )



import geopandas as gpd
import numpy as np
import pytest
import shapely
import xarray as xr

from xugrid import conversion as cv


@pytest.fixture(scope="function")
def line():
    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    edge_node_connectivity = np.array(
        [
            [0, 1],
            [1, 2],
        ]
    )
    return x, y, edge_node_connectivity


@pytest.fixture(scope="function")
def line_gdf():
    x = np.array([0.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 0.0])
    gdf = gpd.GeoDataFrame(geometry=[shapely.linestrings(x, y)])
    return gdf


@pytest.fixture(scope="function")
def triangle_mesh():
    x = np.array([0.0, 1.0, 1.0, 2.0])
    y = np.array([0.0, 0.0, 1.0, 0.0])
    fill_value = -1
    # Two triangles
    faces = np.array(
        [
            [0, 1, 2],
            [1, 3, 2],
        ]
    )
    return x, y, faces, fill_value


@pytest.fixture(scope="function")
def mixed_mesh():
    x = np.array([0.0, 1.0, 1.0, 2.0, 2.0])
    y = np.array([0.0, 0.0, 1.0, 0.0, 1.0])
    fill_value = -1
    # Triangle, quadrangle
    faces = np.array(
        [
            [0, 1, 2, fill_value],
            [1, 3, 4, 2],
        ]
    )
    return x, y, faces, fill_value


@pytest.fixture(scope="function")
def structured_mesh_ascending():
    da = xr.DataArray(
        data=np.arange(12).reshape((3, 4)),
        coords={"y": [5.0, 10.0, 15.0], "x": [2.0, 4.0, 6.0, 8.0]},
        dims=["y", "x"],
        name="grid",
    )
    return da


@pytest.fixture(scope="function")
def structured_mesh_descending():
    da = xr.DataArray(
        data=np.arange(12).reshape((3, 4)),
        coords={"y": [15.0, 10.0, 5.0], "x": [8.0, 6.0, 4.0, 2.0]},
        dims=["y", "x"],
        name="grid",
    )
    return da


def test_nodes_geos_roundtrip(line):
    x, y, _ = line
    actual = cv.nodes_to_points(x, y)
    x_back, y_back = cv.points_to_nodes(actual)
    points_back = cv.nodes_to_points(x_back, y_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(actual, points_back)


def test_linestrings_to_edges(line_gdf):
    x, y, segments = cv.linestrings_to_edges(line_gdf.geometry.values)
    assert np.allclose(x, [0.0, 1.0, 2.0])
    assert np.allclose(y, [0.0, 0.0, 0.0])
    assert np.array_equal(segments, [[0, 1], [1, 2]])


def test_edges_geos_roundtrip(line):
    x, y, c = line
    actual = cv.edges_to_linestrings(x, y, c)
    x_back, y_back, c_back = cv.linestrings_to_edges(actual)
    lines_back = cv.edges_to_linestrings(x_back, y_back, c_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(c, c_back)
    assert np.array_equal(actual, lines_back)


# Cannot use fixtures in parametrize:
# https://github.com/pytest-dev/pytest/issues/349
def _faces_geos_roundtrip(mesh):
    x, y, c, fv = mesh
    actual = cv.faces_to_polygons(x, y, c, fv)
    x_back, y_back, c_back, fv_back = cv.polygons_to_faces(actual)
    polygons_back = cv.faces_to_polygons(x_back, y_back, c_back, fv_back)
    assert np.array_equal(x, x_back)
    assert np.array_equal(y, y_back)
    assert np.array_equal(c, c_back)
    assert fv == fv_back
    assert np.array_equal(actual, polygons_back)


def test_faces_geos_roundtrip__triangle(triangle_mesh):
    _faces_geos_roundtrip(triangle_mesh)


def test_faces_geos_roundtrip__mixed(mixed_mesh):
    _faces_geos_roundtrip(mixed_mesh)


def test_scalar_spacing(structured_mesh_ascending, structured_mesh_descending):
    upcoords = structured_mesh_ascending.coords["x"]
    downcoords = structured_mesh_descending.coords["x"]
    spacing = xr.DataArray(0.1, name="dx")
    with pytest.raises(ValueError, match="spacing of x does not match value of dx"):
        cv._scalar_spacing(upcoords, spacing)

    spacing = xr.DataArray(2.0, name="dx")
    assert np.allclose(cv._scalar_spacing(upcoords, spacing), 1.0)
    spacing = xr.DataArray(-2.0, name="dx")
    assert np.allclose(cv._scalar_spacing(downcoords, spacing), 1.0)


def test_array_spacing(structured_mesh_ascending, structured_mesh_descending):
    upcoords = structured_mesh_ascending.coords["x"]
    downcoords = structured_mesh_descending.coords["x"]
    spacing = xr.DataArray([0.1], name="dx")
    with pytest.raises(ValueError, match="size of x does not match size of dx"):
        cv._array_spacing(upcoords, spacing)

    spacing = xr.DataArray([2.0, 2.0, 2.0, 2.0], name="dx")
    assert np.allclose(cv._array_spacing(upcoords, spacing), 1.0)
    spacing = xr.DataArray([-2.0, -2.0, -2.0, -2.0], name="dx")
    assert np.allclose(cv._array_spacing(downcoords, spacing), 1.0)


def test_implicit_spacing(structured_mesh_ascending, structured_mesh_descending):
    da = xr.DataArray(
        [[0.0, 0.0]],
        {"y": [0.0], "x": [1.0, 2.0]},
        ["y", "x"],
    )
    with pytest.raises(ValueError, match="Cannot derive spacing of 1-sized coordinate"):
        cv._implicit_spacing(da.coords["y"])

    upcoords = structured_mesh_ascending.coords["x"]
    downcoords = structured_mesh_descending.coords["x"]
    assert np.allclose(cv._implicit_spacing(upcoords), 1.0)
    assert np.allclose(cv._implicit_spacing(downcoords), 1.0)


@pytest.mark.parametrize("spacing_type", ["implicit", "scalar", "array"])
def test_infer_bounds(
    structured_mesh_ascending, structured_mesh_descending, spacing_type
):
    up = structured_mesh_ascending
    down = structured_mesh_descending
    x_expected = np.array(
        [
            [1.0, 3.0],
            [3.0, 5.0],
            [5.0, 7.0],
            [7.0, 9.0],
        ]
    )
    y_expected = np.array(
        [
            [2.5, 7.5],
            [7.5, 12.5],
            [12.5, 17.5],
        ]
    )

    if spacing_type == "scalar":
        up = up.assign_coords({"dx": 2.0, "dy": 5.0})
        down = down.assign_coords({"dx": 2.0, "dy": 5.0})
    elif spacing_type == "array":
        up = up.assign_coords({"dx": ("x", [2.0] * 4), "dy": ("y", [5.0] * 3)})
        down = down.assign_coords({"dx": ("x", [2.0] * 4), "dy": ("y", [5.0] * 3)})

    assert np.allclose(cv.infer_bounds(up, "x"), x_expected)
    assert np.allclose(cv.infer_bounds(up, "y"), y_expected)
    assert np.allclose(cv.infer_bounds(down, "x"), x_expected[::-1])
    assert np.allclose(cv.infer_bounds(down, "y"), y_expected[::-1])


def test_infer_bounds_errors(structured_mesh_ascending):
    up = structured_mesh_ascending
    up = up.assign_coords(x=[2.0, 4.0, 3.0, 8.0])

    with pytest.raises(ValueError, match="x is not monotonic"):
        cv.infer_bounds(up, "x")


def test_bounds_to_vertices(structured_mesh_ascending, structured_mesh_descending):
    with pytest.raises(ValueError, match="Bounds are not monotonic"):
        cv.bounds_to_vertices(
            xr.DataArray(
                data=[[0.0, 1.0], [2.0, 3.0], [1.0, 2.0]], dims=["x", "nbound"]
            )
        )

    up = structured_mesh_ascending
    down = structured_mesh_descending
    x_vertices = np.array([1.0, 3.0, 5.0, 7.0, 9.0])
    y_vertices = np.array([2.5, 7.5, 12.5, 17.5])
    # Ascending
    x_bounds = cv.infer_bounds(up, "x")
    y_bounds = cv.infer_bounds(up, "y")
    assert np.allclose(cv.bounds_to_vertices(x_bounds), x_vertices)
    assert np.allclose(cv.bounds_to_vertices(y_bounds), y_vertices)
    # Descending
    x_bounds = cv.infer_bounds(down, "x")
    y_bounds = cv.infer_bounds(down, "y")
    assert np.allclose(cv.bounds_to_vertices(x_bounds), x_vertices[::-1])
    assert np.allclose(cv.bounds_to_vertices(y_bounds), y_vertices[::-1])


def test_infer_xy_coords():
    da = xr.DataArray(
        data=[[1]],
        coords={"y": [1], "x": [1]},
        dims=["y", "x"],
    )
    assert cv.infer_xy_coords(da) == ("x", "y")

    da = xr.DataArray(
        data=[[1]],
        coords={"latitude": [1], "longitude": [1]},
        dims=["latitude", "longitude"],
    )
    assert cv.infer_xy_coords(da) == ("longitude", "latitude")

    da = xr.DataArray(
        data=[[1]],
        coords={"lat": [1], "lon": [1]},
        dims=["lat", "lon"],
    )
    assert cv.infer_xy_coords(da) == (None, None)

    da["lon"].attrs["axis"] = "X"
    da["lat"].attrs["axis"] = "Y"

    da = xr.DataArray(
        data=[[1]],
        coords={"lat": [1], "lon": [1]},
        dims=["lat", "lon"],
    )
    da["lon"].attrs["standard_name"] = "longitude"
    da["lat"].attrs["standard_name"] = "latitude"
    assert cv.infer_xy_coords(da) == ("lon", "lat")

    da = xr.DataArray(
        data=[[1]],
        coords={"yy": [1], "xx": [1]},
        dims=["yy", "xx"],
    )
    da["xx"].attrs["standard_name"] = "projection_x_coordinate"
    da["yy"].attrs["standard_name"] = "projection_y_coordinate"
    assert cv.infer_xy_coords(da) == ("xx", "yy")



import numpy as np
import pytest

from xugrid.regrid.structured import StructuredGrid1d, StructuredGrid2d

# Testgrids
# --------
# grid a(x):               |______50_____|_____100_____|_____150_____|               -> source
# grid b(x):        |______25_____|______75_____|_____125_____|_____175_____|        -> target
# --------
# grid c(x):            |______40_____|______90_____|_____140_____|____190_____|     -> target
# --------
# grid d(x):              |__30__|__55__|__80_|__105__|                              -> target
# --------
# grid e(x):              |__30__|____67.5____|__105__|                              -> target
# --------


def test_init_1d(grid_data_a_1d):
    assert isinstance(grid_data_a_1d, StructuredGrid1d)
    with pytest.raises(TypeError):
        StructuredGrid1d(1)


def test_init_2d(grid_data_a_2d):
    assert isinstance(grid_data_a_2d, StructuredGrid2d)
    with pytest.raises(TypeError):
        StructuredGrid2d(1)


def test_overlap_1d(
    grid_data_a_1d, grid_data_b_1d, grid_data_b_flipped_1d, grid_data_e_1d
):
    # --------
    # source   targets  weight
    # node 0   0, 1     25 m
    # node 1   1, 2     25 m
    # node 2   2, 3     25 m
    # --------
    source, target, weights = grid_data_a_1d.overlap(grid_data_b_1d, relative=False)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 0, 1, 1, 2, 2]))
    assert np.array_equal(target[sorter], np.array([0, 1, 1, 2, 2, 3]))
    assert np.array_equal(weights[sorter], np.array([25, 25, 25, 25, 25, 25]))

    # flipped axis (y-axis)
    # --------
    # source   targets  weight
    # 0        0, 1     25 m
    # 1        1, 2     25 m
    # 2        2, 3     25 m
    # --------
    source, target, weights = grid_data_a_1d.overlap(
        grid_data_b_flipped_1d, relative=False
    )
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 0, 1, 1, 2, 2]))
    assert np.array_equal(target[sorter], np.array([2, 3, 1, 2, 0, 1]))
    assert np.array_equal(weights[sorter], np.array([25, 25, 25, 25, 25, 25]))

    # non-equidistant
    # --------
    # source   targets  weight
    # node 0   0, 1     17.5 m, 32.5 m
    # node 1   1, 2     17.5 m, 25.0 m
    # --------
    source, target, weights = grid_data_a_1d.overlap(grid_data_e_1d, relative=False)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 0, 1, 1]))
    assert np.array_equal(target[sorter], np.array([0, 1, 1, 2]))
    assert np.array_equal(weights[sorter], np.array([17.5, 32.5, 17.5, 25.0]))

    # relative
    # --------
    # source   targets  weight
    # node 0   0, 1     17.5 m, 32.5 m
    # node 1   1, 2     17.5 m, 25.0 m
    # --------
    source, target, weights = grid_data_a_1d.overlap(grid_data_e_1d, relative=True)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 0, 1, 1]))
    assert np.array_equal(target[sorter], np.array([0, 1, 1, 2]))
    assert np.array_equal(
        weights[sorter], np.array([17.5 / 50.0, 32.5 / 50.0, 17.5 / 50.0, 25.0 / 50.0])
    )


def test_overlap_2d(grid_data_a_2d, grid_data_b_2d):
    # --------
    # source   targets            weights
    # 0        0,   1,  4,  5     625 m
    # 1        1,   2,  5,  6     625 m
    # 2        2,   3,  6,  7     625 m
    # 3        4,   5,  8,  9     625 m
    # 4        5,   6,  9, 10     625 m
    # 5        6,   7, 10, 11     625 m
    # 6        8,   9, 12, 13     625 m
    # 7        9,  10, 13, 14     625 m
    # 8        10, 11, 14, 15     625 m
    # --------
    source, target, weights = grid_data_a_2d.overlap(grid_data_b_2d, relative=False)
    sorter = np.argsort(source)
    assert np.array_equal(
        source[sorter],
        np.array(
            [
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                2,
                2,
                2,
                2,
                3,
                3,
                3,
                3,
                4,
                4,
                4,
                4,
                5,
                5,
                5,
                5,
                6,
                6,
                6,
                6,
                7,
                7,
                7,
                7,
                8,
                8,
                8,
                8,
            ]
        ),
    )
    assert np.array_equal(
        target[sorter],
        np.array(
            [
                0,
                4,
                5,
                1,
                2,
                6,
                5,
                1,
                2,
                3,
                7,
                6,
                8,
                9,
                5,
                4,
                9,
                5,
                10,
                6,
                10,
                11,
                7,
                6,
                9,
                8,
                12,
                13,
                10,
                14,
                13,
                9,
                10,
                11,
                14,
                15,
            ]
        ),
    )
    assert np.array_equal(weights[sorter], np.array([625] * source.size))


def test_locate_centroids_1d(
    grid_data_a_1d, grid_data_b_1d, grid_data_b_flipped_1d, grid_data_e_1d
):
    # --------
    # source   target  weight
    # 0        1       1
    # 1        2       1
    # --------
    source, target, weights = grid_data_a_1d.locate_centroids(grid_data_b_1d)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 1]))
    assert np.array_equal(target[sorter], np.array([1, 2]))
    assert np.allclose(weights[sorter], np.ones(2))

    # flipped axis (y-axis)
    # --------
    # source   target  weight
    # 0        1       1
    # 1        2       1
    # --------
    source, target, weights = grid_data_a_1d.locate_centroids(grid_data_b_flipped_1d)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 1]))
    assert np.array_equal(target[sorter], np.array([2, 1]))
    assert np.allclose(weights[sorter], np.ones(2))

    # non-equidistant
    # --------
    # source   target  weight
    # 0        0, 1    1, 1
    # 1        2       1
    # --------
    source, target, weights = grid_data_a_1d.locate_centroids(grid_data_e_1d)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 0, 1]))
    assert np.array_equal(target[sorter], np.array([0, 1, 2]))
    assert np.allclose(weights[sorter], np.ones(3))


def test_locate_centroids_2d(grid_data_a_2d, grid_data_b_2d):
    # --------
    # source   target  weight
    # 0        5       1
    # 1        6       1
    # 3        9       1
    # 4        10      1
    # --------
    source, target, weights = grid_data_a_2d.locate_centroids(grid_data_b_2d)
    sorter = np.argsort(source)
    assert np.array_equal(source[sorter], np.array([0, 1, 3, 4]))
    assert np.array_equal(target[sorter], np.array([5, 6, 9, 10]))
    assert np.allclose(weights[sorter], np.ones(4))


def test_linear_weights_1d(
    grid_data_a_1d,
    grid_data_b_1d,
    grid_data_b_flipped_1d,
    grid_data_c_1d,
    grid_data_d_1d,
    grid_data_e_1d,
):
    # --------
    # source   target  weight
    # 0   ->   1       50%
    # 1   ->   1       50%
    # 1   ->   2       50%
    # 2   ->   2       50%
    # --------
    source, target, weights = grid_data_a_1d.linear_weights(grid_data_b_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([0, 1, 1, 2]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2]))
    assert np.allclose(weights[sorter], np.array([0.5, 0.5, 0.5, 0.5]))

    # flipped axis (y-axis)
    # --------
    # source   target  weight
    # 0   ->   2       50%
    # 1   ->   2       50%
    # 1   ->   1       50%
    # 2   ->   1       50%
    # --------
    source, target, weights = grid_data_a_1d.linear_weights(grid_data_b_flipped_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([1, 2, 0, 1]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2]))
    assert np.allclose(weights[sorter], np.array([0.5, 0.5, 0.5, 0.5]))

    # --------
    # source   target  weight
    # 1        1       80%
    # 0        1       20%
    # 2        2       80%
    # 1        2       20%
    # --------
    source, target, weights = grid_data_a_1d.linear_weights(grid_data_c_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([1, 0, 2, 1]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2]))
    assert np.allclose(weights[sorter], np.array([0.8, 0.2, 0.8, 0.2]))

    # --------
    # source   target  weight
    # 0        1       90%
    # 1        1       10%
    # 0        2       40%
    # 1        2       60% *reversed in output
    # 1        3       90%
    # 2        3       10%
    # --------
    source, target, weights = grid_data_a_1d.linear_weights(grid_data_d_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([0, 1, 1, 0, 1, 2]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2, 3, 3]))
    assert np.allclose(weights[sorter], np.array([0.9, 0.1, 0.6, 0.4, 0.9, 0.1]))

    # non-equidistant
    # --------
    # source   target  weight
    # 0        1       65%
    # 1        1       35%
    # 1        2       90%
    # 2        2       10%
    # --------
    source, target, weights = grid_data_a_1d.linear_weights(grid_data_e_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([0, 1, 1, 2]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2]))
    assert np.allclose(weights[sorter], np.array([0.65, 0.35, 0.9, 0.1]))

    # 1-1 grid
    # --------
    # source   target  weight
    # 1        1       100%
    # 0        1       0%
    # 2        2       100%
    # 1        2       0%
    # --------
    source, target, weights = grid_data_b_1d.linear_weights(grid_data_b_1d)
    sorter = np.argsort(target)
    assert np.array_equal(source[sorter], np.array([1, 0, 2, 1]))
    assert np.array_equal(target[sorter], np.array([1, 1, 2, 2]))
    assert np.allclose(weights[sorter], np.array([1.0, 0.0, 1.0, 0.0]))


def test_linear_weights_2d(
    grid_data_a_2d, grid_data_a_layered_2d, grid_data_b_2d, grid_data_c_2d
):
    # --------
    # source   targets     weight
    # 5        0, 1, 3, 4  25%
    # 6        1, 2, 4, 5  25%
    # 9        3, 4, 6, 7  25%
    # 10       4, 5, 7, 8  25%
    # --------
    source, target, weights = grid_data_a_2d.linear_weights(grid_data_b_2d)
    sorter = np.argsort(target)
    assert np.array_equal(
        source[sorter], np.array([0, 1, 3, 4, 1, 2, 4, 5, 3, 4, 6, 7, 4, 5, 7, 8])
    )
    assert np.array_equal(
        target[sorter], np.array([5, 5, 5, 5, 6, 6, 6, 6, 9, 9, 9, 9, 10, 10, 10, 10])
    )
    assert np.allclose(weights[sorter], np.array([0.25] * 16))

    # --------
    # source   targets      weight
    # 5        0, 1, 3, 4   10%	40%	10%	40%
    # 6        1, 2, 4, 5   10%	40%	10%	40%
    # 9        3, 4, 6, 7   10%	40%	10%	40%
    # 10       4, 5, 7, 8   10%	40%	10%	40%
    # --------
    source, target, weights = grid_data_a_layered_2d.linear_weights(grid_data_c_2d)
    sorter = np.argsort(target)
    assert np.array_equal(
        source[sorter], np.array([1, 0, 4, 3, 2, 1, 5, 4, 4, 3, 7, 6, 5, 4, 8, 7])
    )
    assert np.array_equal(
        target[sorter], np.array([5, 5, 5, 5, 6, 6, 6, 6, 9, 9, 9, 9, 10, 10, 10, 10])
    )
    assert np.allclose(weights[sorter], np.array([0.4, 0.1, 0.4, 0.1] * 4))

    # 1-1
    # --------
    # source   targets      weight
    # 5        4, 5, 8, 9   0% 100% 0% 0%
    # 6        5, 6, 9,10   0% 100% 0% 0%
    # 9        8, 9,12,13   0% 100% 0% 0%
    # 10       9,10,13,14   0% 100% 0% 0%
    # --------
    source, target, weights = grid_data_b_2d.linear_weights(grid_data_b_2d)
    sorter = np.argsort(target)
    assert np.array_equal(
        source[sorter], np.array([5, 4, 9, 8, 6, 5, 10, 9, 9, 8, 13, 12, 10, 9, 14, 13])
    )
    assert np.array_equal(
        target[sorter], np.array([5, 5, 5, 5, 6, 6, 6, 6, 9, 9, 9, 9, 10, 10, 10, 10])
    )
    assert np.allclose(
        weights[sorter], np.array([1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0])
    )
